@@ -12,8 +12,6 @@
 #
 ##############################################################################
 """Buildout package
-
-$Id$
 """
 
 class UserError(Exception):
@@ -21,4 +19,4 @@ class UserError(Exception):
     """
 
     def __str__(self):
-        return " ".join(map(str, self))
+        return " ".join(map(str, self.args))
