@@ -20,19 +20,32 @@ except ImportError:
 
 try:
     # Python 3
-    from urllib.request import FancyURLopener, URLopener, urlretrieve
+    from urllib.request import urlretrieve
     from urllib.parse import urlparse
-    from urllib import request as urllib # for monkey patch below :(
 except ImportError:
     # Python 2
-    from urllib import FancyURLopener, URLopener, urlretrieve
+    import base64
     from urlparse import urlparse
-    import urllib
+    from urlparse import urlunparse
+    import urllib2
 
-class URLOpener(FancyURLopener):
-    http_error_default = URLopener.http_error_default
-
-urllib._urlopener = URLOpener() # Ook! Monkey patch!
+    def urlretrieve(url, tmp_path):
+        """Work around Python issue 24599 includig basic auth support
+        """
+        scheme, netloc, path, params, query, frag = urlparse(url)
+        auth, host = urllib2.splituser(netloc)
+        if auth:
+            url = urlunparse((scheme, host, path, params, query, frag))
+            req = urllib2.Request(url)
+            base64string = base64.encodestring(auth)[:-1]
+            basic = "Basic " + base64string
+            req.add_header("Authorization", basic)
+        else:
+            req = urllib2.Request(url)
+        url_obj = urllib2.urlopen(req)
+        with open(tmp_path, 'wb') as fp:
+            fp.write(url_obj.read())
+        return tmp_path, url_obj.info()
 
 
 from zc.buildout.easy_install import realpath
@@ -45,6 +58,7 @@ import sys
 import tempfile
 import zc.buildout
 
+
 class ChecksumError(zc.buildout.UserError):
     pass
 
