@@ -25,6 +25,7 @@ except ImportError:
 
 import errno
 import logging
+from multiprocessing import Process
 import os
 import pkg_resources
 import random
@@ -46,8 +47,9 @@ print_ = zc.buildout.buildout.print_
 fsync = getattr(os, 'fsync', lambda fileno: None)
 is_win32 = sys.platform == 'win32'
 
-setuptools_location = pkg_resources.working_set.find(
-    pkg_resources.Requirement.parse('setuptools')).location
+def read(path='out', *rest):
+    with open(os.path.join(path, *rest)) as f:
+        return f.read()
 
 def cat(dir, *names):
     path = os.path.join(dir, *names)
@@ -56,13 +58,28 @@ def cat(dir, *names):
         and os.path.exists(path+'-script.py')
         ):
         path = path+'-script.py'
-    print_(open(path).read(), end='')
+    with open(path) as f:
+        print_(f.read(), end='')
+
+def eqs(a, *b):
+    a = set(a); b = set(b)
+    return None if a == b else (a - b, b - a)
+
+def clear_here():
+    for name in os.listdir('.'):
+        if os.path.isfile(name) or os.path.islink(name):
+            os.remove(name)
+        else:
+            shutil.rmtree(name)
 
 def ls(dir, *subs):
     if subs:
         dir = os.path.join(dir, *subs)
     names = sorted(os.listdir(dir))
     for name in names:
+        # If we're running under coverage, elide coverage files
+        if os.getenv("COVERAGE_PROCESS_START") and name.startswith('.coverage.'):
+            continue
         if os.path.isdir(os.path.join(dir, name)):
             print_('d ', end=' ')
         elif os.path.islink(os.path.join(dir, name)):
@@ -109,12 +126,19 @@ def clean_up_pyc(*path):
 MUST_CLOSE_FDS = not sys.platform.startswith('win')
 
 def system(command, input='', with_exit_code=False):
+    # Some TERMinals, expecially xterm and its variants, add invisible control
+    # characters, which we do not want as they mess up doctests.  See:
+    # https://github.com/buildout/buildout/pull/311
+    # http://bugs.python.org/issue19884
+    env = dict(os.environ, TERM='dumb')
+
     p = subprocess.Popen(command,
                          shell=True,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
-                         close_fds=MUST_CLOSE_FDS)
+                         close_fds=MUST_CLOSE_FDS,
+                         env=env)
     i, o, e = (p.stdin, p.stdout, p.stderr)
     if input:
         i.write(input.encode())
@@ -127,6 +151,7 @@ def system(command, input='', with_exit_code=False):
         # Use the with_exit_code=True parameter when you want to test the exit
         # code of the command you're running.
         output += 'EXIT CODE: %s' % p.wait()
+    p.wait()
     return output
 
 def get(url):
@@ -142,7 +167,10 @@ def _runsetup(setup, *args):
         os.chdir(os.path.dirname(setup))
         zc.buildout.easy_install.call_subprocess(
             [sys.executable, setup] + args,
-            env=dict(os.environ, PYTHONPATH=setuptools_location))
+            env=dict(os.environ,
+                     PYTHONPATH=zc.buildout.easy_install.setuptools_pythonpath,
+                     ),
+            )
         if os.path.exists('build'):
             rmtree('build')
     finally:
@@ -174,14 +202,21 @@ def wait_until(label, func, *args, **kw):
 
 class TestOptions(zc.buildout.buildout.Options):
 
+    def __init__(self, *args):
+        zc.buildout.buildout.Options.__init__(self, *args)
+        self._created = []
+
     def initialize(self):
         pass
 
 class Buildout(zc.buildout.buildout.Buildout):
 
     def __init__(self):
+        for name in 'eggs', 'parts':
+            if not os.path.exists(name):
+                os.mkdir(name)
         zc.buildout.buildout.Buildout.__init__(
-            self, '', [('buildout', 'directory', os.getcwd())])
+            self, '', [('buildout', 'directory', os.getcwd())], False)
 
     Options = TestOptions
 
@@ -205,6 +240,9 @@ def buildoutSetUp(test):
             root_logger.removeHandler(handler)
         for handler in handlers_before_set_up:
             root_logger.addHandler(handler)
+        bo_logger = logging.getLogger('zc.buildout')
+        for handler in bo_logger.handlers[:]:
+            bo_logger.removeHandler(handler)
     register_teardown(restore_root_logger_handlers)
 
     base = tempfile.mkdtemp('buildoutSetUp')
@@ -240,9 +278,8 @@ def buildoutSetUp(test):
     os.chdir(sample)
 
     # Create a basic buildout.cfg to avoid a warning from buildout:
-    open('buildout.cfg', 'w').write(
-        "[buildout]\nparts =\n"
-        )
+    with open('buildout.cfg', 'w') as f:
+        f.write("[buildout]\nparts =\n")
 
     # Use the buildout bootstrap command to create a buildout
     zc.buildout.buildout.Buildout(
@@ -260,6 +297,38 @@ def buildoutSetUp(test):
     # way due to the trick above:
     os.mkdir('develop-eggs')
 
+    if os.getenv("COVERAGE_PROCESS_START"):
+        # The user has requested subprocess code coverage. Since we will be changing
+        # directories, we need to make sure this path is absolute, which means
+        # we need to temporarily return to our starting directory.
+        os.chdir(here)
+        path_to_coveragerc = os.path.abspath(os.environ['COVERAGE_PROCESS_START'])
+        os.chdir(sample)
+        assert os.path.isfile(path_to_coveragerc), path_to_coveragerc
+        os.environ['COVERAGE_PROCESS_START'] = path_to_coveragerc
+
+        # Before we return to the current directory and destroy the
+        # temporary working directory, we need to copy all the coverage files
+        # back so that they can be `coverage combine`d.
+        def copy_coverage_files():
+            coveragedir = os.path.dirname(path_to_coveragerc)
+            import glob
+            for f in glob.glob('.coverage*'):
+                shutil.copy(f, coveragedir)
+        __tear_downs.insert(0, copy_coverage_files)
+
+
+        # Now we must modify the newly created bin/buildout to
+        # actually begin coverage.
+        with open('bin/buildout') as f:
+            import textwrap
+            lines = f.read().splitlines()
+            assert lines[1] == '', lines
+            lines[1] = 'import coverage; coverage.process_startup()'
+
+        with open('bin/buildout', 'w') as f:
+            f.write('\n'.join(lines))
+
     def start_server(path):
         port, thread = _start_server(path, name=path)
         url = 'http://localhost:%s/' % port
@@ -291,6 +360,7 @@ def buildoutSetUp(test):
         sdist = sdist,
         bdist_egg = bdist_egg,
         start_server = start_server,
+        stop_server = stop_server,
         buildout = os.path.join(sample, 'bin', 'buildout'),
         wait_until = wait_until,
         print_ = print_,
@@ -328,6 +398,7 @@ class Handler(BaseHTTPRequestHandler):
 
     def do_GET(self):
         if '__stop__' in self.path:
+            self.__server.server_close()
             raise SystemExit
 
         def k():
@@ -375,7 +446,8 @@ class Handler(BaseHTTPRequestHandler):
             self.send_header('Content-Length', str(len(out)))
             self.send_header('Content-Type', 'text/html')
         else:
-            out = open(path, 'rb').read()
+            with open(path, 'rb') as f:
+                out = f.read()
             self.send_header('Content-Length', len(out))
             if path.endswith('.egg'):
                 self.send_header('Content-Type', 'application/zip')
@@ -398,6 +470,7 @@ def _run(tree, port):
     server_address = ('localhost', port)
     httpd = Server(tree, server_address, Handler)
     httpd.serve_forever()
+    httpd.server_close()
 
 def get_port():
     for i in range(10):
@@ -452,7 +525,7 @@ def wait(port, up):
         if up:
             raise
         else:
-            raise SystemError("Couln't stop server")
+            raise SystemError("Couldn't stop server")
 
 def install(project, destination):
     if not isinstance(destination, str):
@@ -471,8 +544,8 @@ def install(project, destination):
             shutil.copyfile(dist.location, destination)
     else:
         # copy link
-        open(os.path.join(destination, project+'.egg-link'), 'w'
-             ).write(dist.location)
+        with open(os.path.join(destination, project+'.egg-link'), 'w') as f:
+            f.write(dist.location)
 
 def install_develop(project, destination):
     if not isinstance(destination, str):
@@ -481,8 +554,8 @@ def install_develop(project, destination):
 
     dist = pkg_resources.working_set.find(
         pkg_resources.Requirement.parse(project))
-    open(os.path.join(destination, project+'.egg-link'), 'w'
-         ).write(dist.location)
+    with open(os.path.join(destination, project+'.egg-link'), 'w') as f:
+        f.write(dist.location)
 
 def _normalize_path(match):
     path = match.group(1)
@@ -510,10 +583,10 @@ if sys.version_info > (2, ):
         re.compile('(\n?)d  __pycache__\n'), '\\1')
 else:
     normalize___pycache__ = (
-        re.compile('(\n?)-  \S+\.pyc\n'), '\\1')
+        re.compile(r'(\n?)-  \S+\.pyc\n'), '\\1')
 
 normalize_egg_py = (
-    re.compile('-py\d[.]\d(-\S+)?.egg'),
+    re.compile(r'-py\d[.]\d(-\S+)?.egg'),
     '-pyN.N.egg',
     )
 
@@ -521,6 +594,9 @@ normalize_exception_type_for_python_2_and_3 = (
     re.compile(r'^(\w+\.)*([A-Z][A-Za-z0-9]+Error: )'),
     '\2')
 
+normalize_open_in_generated_script = (
+    re.compile(r"open\(__file__, 'U'\)"), 'open(__file__)')
+
 not_found = (re.compile(r'Not found: [^\n]+/(\w|\.)+/\r?\n'), '')
 
 # Setuptools now pulls in dependencies when installed.
@@ -531,3 +607,37 @@ ignore_not_upgrading = (
     re.compile(
     'Not upgrading because not running a local buildout command.\n'
     ), '')
+
+def run_buildout(command):
+    # Make sure we don't get .buildout
+    os.environ['HOME'] = os.path.join(os.getcwd(), 'home')
+    args = command.split()
+    import pkg_resources
+    buildout = pkg_resources.load_entry_point(
+        'zc.buildout', 'console_scripts', args[0])
+    buildout(args[1:])
+
+def run_from_process(target, *args, **kw):
+    sys.stdout = sys.stderr = open('out', 'w')
+    target(*args, **kw)
+
+def run_in_process(*args, **kwargs):
+    process = Process(target=run_from_process, args=args, kwargs=kwargs)
+    process.daemon = True
+    process.start()
+    process.join(99)
+    if process.is_alive() or process.exitcode:
+        with open('out') as f:
+            print(f.read())
+
+def run_buildout_in_process(command='buildout'):
+    command = command.split(' ', 1)
+    command.insert(
+        1,
+        " use-dependency-links=false"
+        # Leaving this here so we can uncomment to see what's going on.
+        #" log-format=%(asctime)s____%(levelname)s_%(message)s -vvv"
+        " index=" + __file__ + 'nonexistent' # hide index
+        )
+    command = ' '.join(command)
+    run_in_process(run_buildout, command)
